//	CrystalApp.swift
//
//	© 2024 by Jeff Weeks
//	See TermsOfUse.txt

import SwiftUI


@main
struct CrystalApp: App {

	@AppStorage("Crystal") var crystal: Crystal = .iceHexagonal

	//	An ångström is 10⁻¹⁰ meters so if, for example, we use
	//	a scale factor of 10⁹, then an atom whose true radius
	//	is 1 ångström will appear in the simimulation
	//	as a sphere of radius 10⁹ * 10⁻¹⁰ meters = 10 centimeters.
	@AppStorage("scale factor") var scaleFactor: Double
							= defaultScaleFactor	//	dimensionless

	//	The user may wish to shrink all atoms' radii,
	//	to show the crystal's large-scale structure more clearly.
	//
	//		In principle atomicRadiusFactor ∈ 0% … 100%,
	//		but in practice we restrict it to 25% … 100%.
	//
	@AppStorage("Atomic radius factor") var atomicRadiusFactor: Double
							= defaultAtomicRadiusFactor	//	dimensionless
	
	@AppStorage("Show bonds") var showBonds: Bool = true


	init() {
		ManifestComponent.registerComponent()
	}

	var body: some Scene {

		WindowGroup(id: "Control panel", for: Int.self) { _ in

			CrystalControlPanel(
				crystal: $crystal,
				scaleFactor: $scaleFactor,
				atomicRadiusFactor: $atomicRadiusFactor,
				showBonds: $showBonds)
				
		} defaultValue: { 0 }	//	assigns value 0 to the initial window
		.defaultSize(width: 512.0, height: 480.0)

		ImmersiveSpace(id: "Immersive crystal") {
			CrystalImmersiveView(
				crystal: crystal,
				scaleFactor: scaleFactor,
				showBonds: showBonds,
				atomicRadiusFactor: $atomicRadiusFactor)
		}
		.immersionStyle(selection: .constant(.full), in: .full)
		.persistentSystemOverlays(.hidden)
			//	Asks visionOS not to interpret palm-up as a request for a home button.
			//	visionOS may, at whim, choose not to comply.

#warning("Delete this temporary alert when no longer needed.")
		WindowGroup(id: "TEMPORARY alert", for: Int.self) { _ in
			TEMPORARYView()
		} defaultValue: { 0 }	//	assigns value 0 to the initial window
		.defaultSize(width: 256.0, height: 336.0)
	}
}

#warning("Delete this TEMPORARYView when no longer needed.")
struct TEMPORARYView: View {

	@Environment(\.dismissWindow) var dismissWindow

	var body: some View {

		VStack(spacing: 16.0) {

			Text("To close the crystal, please\npress the Digital Crown.")
			.fixedSize(horizontal: false, vertical: true)
				//	fixedSize() is needed to allow multiple lines
				//	if the enclosing frame's height isn't specified
				//	(unintuitive but true)
			.multilineTextAlignment(.center)

			ZStack(alignment: .center) {
			
				Color.white
				.frame(width: 144.0, height: 96.0)
				.cornerRadius(16.0)
				
				Image("Press Digital Crown")
				.resizable()
				.frame(width: 105.0, height: 90.0)
			}

			Text("Double-tapping is no longer\nused for this purpose.")
			.font(.caption2)
			.fixedSize(horizontal: false, vertical: true)
				//	fixedSize() is needed to allow multiple lines
				//	if the enclosing frame's height isn't specified
				//	(unintuitive but true)
			.multilineTextAlignment(.center)

			Button("OK") {
				dismissWindow(id: "TEMPORARY alert", value: 0)
			}
		}
	}
}
