//	CrystalControlPanel.swift
//
//	© 2024 by Jeff Weeks
//	See TermsOfUse.txt

import SwiftUI

let minScaleFactor      = 1.0e9  //  dimensionless, 10⁹ Å  =  10 cm
let maxScaleFactor      = 1.0e10 //  dimensionless, 10¹⁰ Å = 100 cm
let defaultScaleFactor  = 2.5e9  //  dimensionless,		   =  25 cm

let minAtomicRadiusFactor		= 0.25  //  dimensionless
let maxAtomicRadiusFactor		= 1.00  //  dimensionless
let defaultAtomicRadiusFactor	= 0.60  //  dimensionless


struct CrystalControlPanel: View {

	@Binding var crystal: Crystal
	@Binding var scaleFactor: Double		//	dimensionless, ∈  10⁹ … 10¹⁰
	@Binding var atomicRadiusFactor: Double	//	dimensionless, ∈  25% … 100%
	@Binding var showBonds: Bool

	@AppStorage("Show instructions") var showInstructions: Bool = true
	@State var instructionsSheetIsPresented: Bool = false

	@State var settingsSheetIsPresented: Bool = false

	@Environment(\.openImmersiveSpace) var openImmersiveSpace


	var body: some View {

		VStack(spacing: 32.0) {
		
			Text("Crystal Explorer")
			.font(.title)
		
			CrystalChoiceView(
				crystal: $crystal,
				instructionsSheetIsPresented: $instructionsSheetIsPresented,
				showInstructions: showInstructions)
			
			Button(
				action: {
					settingsSheetIsPresented = true
				}, label: {
					Label("Settings", systemImage: "gear")
				}
			)
 		}
		.padding(32.0)
		.sheet(
			isPresented: $settingsSheetIsPresented,
			content: {
				SettingsView(
					scaleFactor: $scaleFactor,
					atomicRadiusFactor: $atomicRadiusFactor,
					showBonds: $showBonds,
					showInstructions: $showInstructions)
			}
		)
		.sheet(
			isPresented: $instructionsSheetIsPresented,
			onDismiss: {
				Task { await openImmersiveSpace(id: "Immersive crystal") }
			},
			content: {
				InstructionsView(showInstructions: $showInstructions)
			}
		)
	}
}

struct CrystalChoiceView: View {

	@Binding var crystal: Crystal
	@Binding var instructionsSheetIsPresented: Bool

	var showInstructions: Bool

	@Environment(\.openImmersiveSpace) var openImmersiveSpace

	var body: some View {

		Grid() {
			GridRow() {
				crystalButton(.iceHexagonal)
				crystalButton(.iceCubic)
			}
			GridRow() {
				crystalButton(.halite)
				crystalButton(.fluorite)
			}
			GridRow() {
				crystalButton(.diamond)
				crystalButton(.quartz)
			}
		}
	}
	
	func crystalButton(
		_ proposedCrystal: Crystal
		
	) -> some View {
	
		Button(
			action: {

				crystal = proposedCrystal

				if showInstructions {
					instructionsSheetIsPresented = true
				} else {
					Task { await openImmersiveSpace(id: "Immersive crystal") }
				}
			},
			label: {
			
				HStack() {
				
					Image("Crystal/" + proposedCrystal.formalNameKey)
						.cornerRadius(4.0)
					
					VStack(alignment: .leading) {
					
						Text(proposedCrystal.familiarNameKey)
							.lineLimit(1)
							.foregroundColor(.primary)
							
						Text(verbatim: proposedCrystal.formula)
							.font(.footnote)
							.foregroundColor(.secondary)
					}
					
					Spacer()
				}
			}
		)
	}
}

struct SettingsView: View {

	@Binding var scaleFactor: Double		//	dimensionless, ∈  10⁹ … 10¹⁰
	@Binding var atomicRadiusFactor: Double	//	dimensionless, ∈  25% … 100%
	@Binding var showBonds: Bool
	@Binding var showInstructions: Bool


	@Environment(\.dismiss) var dismiss	//	to dismiss a sheet

	var body: some View {

		VStack(spacing: 32.0) {
		
			DimensionSliderView(
				title: "Scale",
				value: $scaleFactor,
				minValue: minScaleFactor,
				maxValue: maxScaleFactor,
				units: .crystalScaleFactor)

			VStack(spacing: 6.0) {
			
				DimensionSliderView(
					title: "Atom size",
					value: $atomicRadiusFactor,
					minValue: minAtomicRadiusFactor,
					maxValue: maxAtomicRadiusFactor,
					units: .percentage)

				Text("While in the crystal you may use a Magnify\ngesture to change the atom size.")
				.font(.caption2)
				.fixedSize(horizontal: false, vertical: true)
					//	fixedSize() is needed to allow multiple lines
					//	if the enclosing frame's height isn't specified
					//	(unintuitive but true)
				.multilineTextAlignment(.center)	//	the default
			}
			
			Toggle("Show bonds", isOn: $showBonds)
			
			Toggle("Show instructions when opening crystals", isOn: $showInstructions)

			Button("Done") {
				dismiss()
			}
		}
		.padding(32.0)
	}
}

struct InstructionsView: View {

	@Binding var showInstructions: Bool

	@Environment(\.dismiss) var dismiss	//	to dismiss a sheet

	var body: some View {
	
		VStack(spacing: 16.0) {

			Grid(alignment: .leading, horizontalSpacing: 16.0) {
			
				GridRow() {
				
					ZStack(alignment: .center) {
					
						Color.white
						.frame(width: 144.0, height: 96.0)
						.cornerRadius(16.0)
						
						Image("Magnify")
						.resizable()
						.frame(width: 128.0, height: 64.0)
					}
						
					VStack(alignment: .leading, spacing: 8.0) {
					
						Text("Move pinching hands apart")
						.bold()
						
						Text("    to expand or shrink the atoms.")
					}
				}
				
				GridRow() {
				
					ZStack(alignment: .center) {
					
						Color.white
						.frame(width: 144.0, height: 96.0)
						.cornerRadius(16.0)
						
						Image("Press Digital Crown")
						.resizable()
						.frame(width: 105.0, height: 90.0)
					}
						
					VStack(alignment: .leading, spacing: 8.0) {
					
						Text("Press the Digital Crown")
						.bold()
						
						Text("    when you want to close the crystal.")
					}
				}
			}
			
			Button("Got it") {
				dismiss()
			}
						
			Toggle(isOn: $showInstructions)
			{
				Text("Show these instructions when opening crystals")
				.font(.caption)
			}
			.fixedSize()
		}
		.padding(32.0)
	}
}
