//	CrystalCrystals.swift
//
//	© 2024 by Jeff Weeks
//	See TermsOfUse.txt

import simd


enum Element {

	case hydrogen
	case carbon
	case oxygen
	case fluorine
	case sodium
	case silicon
	case chlorine
	case calcium

	var linearDisplayP3Color: SIMD3<Double> {
	
		switch self {
		case .hydrogen:	return SIMD3<Double>( 1.00, 1.00, 1.00 )	//	linear Display P3
		case .carbon: 	return SIMD3<Double>( 0.00, 0.25, 1.00 )
		case .oxygen:	return SIMD3<Double>( 1.00, 0.00, 0.00 )
		case .fluorine:	return SIMD3<Double>( 0.00, 0.75, 1.00 )
		case .sodium:	return SIMD3<Double>( 0.25, 0.00, 1.00 )
		case .silicon:	return SIMD3<Double>( 1.00, 0.75, 0.00 )
		case .chlorine:	return SIMD3<Double>( 0.00, 1.00, 0.50 )
		case .calcium:	return SIMD3<Double>( 0.50, 0.00, 1.00 )
		}
	}
}

enum BondType
{
	case covalent
	case ionic
	case hydrogen
	
	var radius: Double {
	
		let theFullRadius = 0.125	//	in ångströms

		let theRadiusFactor: Double
		switch self {
		case .covalent:
			theRadiusFactor = 1.0
		case .ionic:
			theRadiusFactor = 0.5
		case .hydrogen:
			theRadiusFactor = 0.25
		}
		
		let theRadius = theRadiusFactor * theFullRadius
		
		return theRadius
	}
	
	var linearDisplayP3Color: SIMD3<Double> {
	
		switch self {
		case .covalent:	return SIMD3<Double>(0.870, 0.792, 0.467)	//	brass
		case .ionic:	return SIMD3<Double>(0.912, 0.914, 0.919)	//	aluminum
		case .hydrogen:	return SIMD3<Double>(0.640, 0.651, 0.662)	//	chromium
		}
	}
}

struct Atom {

	let element: Element
	let radius: Double	//	in ångströms, at actual size;
						//		the user-interface code may shrink all atoms
						//		to make the crystal structure more visible
						//	Note: the radius depends not only on what element it is,
						//		but also what ionization state it's in.
	let center: SIMD3<Double>	//	in ångströms
}

struct Bond {

	let type: BondType
	let start: SIMD3<Double>	//	in ångströms, at center of one atom
	let finish: SIMD3<Double>	//	in ångströms, at center of other atom
}

enum Crystal: Int, CaseIterable, Identifiable {

	case iceHexagonal
	case iceCubic
	case halite
	case fluorite
	case diamond
	case quartz
	
	var id: Int { self.rawValue }

	//	Unit cell
	//
	//	This Crystal enum provides one unit cell's worth of Atoms and Bonds
	//	for each Crystal.  A "unit cell" is a polyhedron which, when repeated
	//	using the translational part of the crystal's symmetry group,
	//	tiles the whole crystal with no gaps or overlaps.

	//	Special treatment for quartz
	//
	//	For quartz, the unit cell is a regular hexagonal prism.
	//	This unit cell has an internal symmetry given by a corkscrew motion
	//	along the prism's central axis.  The corkscrew motion translates
	//	a distance of 1/3 the prism's height, while rotating 1/3 of a turn.
	//	Thus 1/3 of the prism determines the whole crystal structure, and
	//	our internal unifiedData defines Atoms and Bonds only for that 1/3 of the prism,
	//	leaving it to the generateQuartzAtoms() and generateQuartzBonds() functions
	//	to replicate those Atoms and Bonds for the other two "1/3 prisms",
	//	using the corkscrew symmetry that out internal CrystalData provides.

	//	Organizational note
	//
	//	While I want this Crystal enum to provide separate properties
	//	for a crystal's atoms, bonds, etc., I prefer to keep all
	//	all of a given crystal's data together in one place.
	//	So I've put the unified CrystalData into a private var below,
	//	and let the separate property variables access that unified data.
	
	var formalNameKey: String {
		return unifiedData.formalNameKey
	}
	
	var familiarNameKey: String {
		return unifiedData.familiarNameKey
	}
	
	var formula: String {
		return unifiedData.formula
	}
	
	var atoms: [Atom] {
	
		if self == .quartz {
			return generateQuartzAtoms()
		} else {
			return unifiedData.atoms
		}
	}
	
	var bonds: [Bond] {
	
		if self == .quartz {
			return generateQuartzBonds()
		} else {
			return unifiedData.bonds
		}
	}

	var primitiveTranslations: simd_double3x3 {
	
		//	The primitiveTranslations generate the translational part
		//	of the crystal's symmetry group (in effect the Bravais lattice).
		//
		//	When written as the columns (resp. rows) of a matrix
		//	using the "matrix times column vector" convention
		//	(resp. "row vector times matrix" convention),
		//	the primitiveTranslations define a mapping
		//	from a unit cubic lattice to the crystal's actual lattice.
		
		return unifiedData.primitiveTranslations
	}

	var aestheticOffset: SIMD3<Double> {	//	User is facing forwards in all cases.

		//	How should we offset the crystal so that the user
		//	sees an aesthetically pleasing view at the first moment?
		//	In particular, we want to make sure that the user
		//	doesn't start with his/her head inside an atom or bond.
		//	We might also include some other aesthetic improvements,
		//	for example in hexagonal ice we might want the user
		//	to start in the center of a hexagonal tube,
		//	looking down that tube.

		return unifiedData.aestheticOffset
	}

	
	struct CrystalData {
		let formalNameKey: String
		let familiarNameKey: String
		let formula: String
		let atoms: [Atom]
		let bonds: [Bond]
		let primitiveTranslations: simd_double3x3	//	in ångströms
		var corkscrew: simd_double4x4? = nil	//	used only for quartz
		let aestheticOffset: SIMD3<Double>
	}
	
	private var unifiedData: CrystalData {

		//	Note on atomic radii
		//
		//	While the length of a bond is a well defined concept
		//	(it's the average distance between the two atoms' nuclei)
		//	and can be measured experimentally (by x-ray diffraction),
		//	the radius of an atom is less well defined,
		//	because the atom's tiny nucleus is surrounded
		//	by a diffuse electron with no well defined boundary.
		//	Even though people often draw the atoms in a crystal
		//	as solid balls packed tightly together, this is
		//	only for ease of illustration, and isn't intended
		//	as a definitive picture of the underlying physical reality.
		//
		//	To create a solid-balls-packed-together picture,
		//	for example for a halite (NaCl) crystal,
		//	we're given the bond length as 2.8201 Å,
		//	but we must then decide how much of that 2.8201 Å
		//	to allocate to the Na⁺ ion and how much to the Cl⁻.
		//	Given the halite example alone, there's no reasonable way
		//	to do that.  But people have measured the bond lengths
		//	in thousands of crystals, in which various ions participate
		//	in various combinations (KCl, CaCl, NaF, etc.) and then
		//	(I'm guessing) have done some sort of least-squares fit
		//	to determine how much, on average, each atom or ion
		//	contributes to the bonds in which it participates.
		//	One can find tables of such atomic/covalent/ionic radii online.
		//	They should, however, be used with caution
		//	("be taken with a grain of salt"): in any particular crystal,
		//	the bond length won't be exactly the sum of the two published radii,
		//	because each published radius is merely the average value
		//	over the many bonds in which that particular atom or ion participates.

		//	Presentation note
		//
		//	If we draw covalently or ionically bounded atoms
		//	as perfectly tangent spheres, the white tubes representing
		//	the bonds would remain slightly visible.
		//	So instead let's enlarge the atoms slightly,
		//	to hide the covalent and ionic bond tubes
		//	when the atoms are drawn at 100% size.
		//
		//		Note:  The atoms are drawn not as perfect spheres,
		//		but rather as polyhedral approximations inscribed
		//		within the intended spheres.  The  enlargement factor
		//		must be large enough to hide the white tubes
		//		even when drawing the atoms at their lowest level of detail.
		//		(This effect is most noticable for the diamond crystal.)
		//
		//	As an added bonus, the overlapping spheres might subtly encourage
		//	the idea that atoms aren't hard spheres like pool balls.
		//
		let theAtomEnlargementFactor = 1.08
		
		let root2: Double = sqrt(2.0)
		let root3: Double = sqrt(3.0)

		//	The following comments and constants apply
		//	to both hexagonal ice and cubic ice
		//	(that's why they're defined here, instead of inside
		//	the .iceHexagonal or .iceCubic cases in the switch
		//	that follows).
		//
		//	As explained in the "Note on atomic radii" above,
		//	it's conventional (although not physically accurate)
		//	to draw atoms as spheres that just barely touch each other.
		//	In ice, this convention gets a little tricky,
		//	because each hydrogen atom sits 1.01 Å away
		//	from one neighboring oxygen atom (the covalently bonded one)
		//	but 1.76 Å away from its other neighbor (the hydrogen-bonded one).
		//	Clearly the sphere representing the hydrogen can't be
		//	tangent to *both* oxygen spheres.
		//	The most satisfactory image results if we let the hydrogen sphere
		//	get "absorbed into" the covalently bonded oxygen sphere
		//	while just barely touching the hydrogen-bonded oxygen sphere.
		//	This gives a set of obvious water molecules, which are
		//	exactly tangent to each other at each hydrogen bond.
		//
		//	Confession:  I couldn't find any authoritative values
		//	for such oxygen and hydrogen radii online, so I simply
		//	made up values that looked good.  The important thing
		//	is that oxygen and hydrogen radii sum to the length
		//	of the hydrogen bond:
		//
		//				1.10 + 0.65 = 2.76 - 1.01
		//
		//	Assume inter-atomic distances will be almost the same
		//	in hexagonal and cubic ice.
		let theIceOxygenRadius = 1.10 // in ångströms
		let theIceHydrogenRadius = 0.65 // in ångströms
		let theIceOO = 2.76 // = O-O distance in hexagonal ice Ih (and also cubic ice Ic?), in ångströms
		let theIceCovalentBondLength = 1.01 // in ångströms (source: https://msu.edu/course/css/850/snapshot.afs/teppen/physical_chemistry_of_water.htm )
		let theIceHydrogenBondLength = theIceOO - theIceCovalentBondLength // in ångströms
		let theIceCovalentBondFraction = theIceCovalentBondLength / theIceOO


		switch self {
		
		case .iceHexagonal:
		
			//	The four bonds (two covalent bonds, two hydrogen bonds) attached
			//	to each water molecule meet with roughly tetrahedral symmetry.
			//	The exact angle θ at which any two of them would meet could,
			//	if the symmetry were perfect, be deduced by considering
			//	a tiling of a unit sphere by (π/2, π/3, π/3) spherical triangles.
			//	In particular, the desired angle ­θ is twice the angle subtended
			//	by a leg of a (π/2, π/3, π/3) spherical triangle.
			//	The spherical law of cosines for angles say
			//
			//		cos γ = -cos α cos β + sin α sin β cos c
			//	 ⇒
			//		cos π/3 = - cos π/2 cos π/3 + sin π/2 sin π/3 cos θ/2
			//	 ⇒
			//		1/2 = 0 + (√3 / 2) cos θ/2
			//	 ⇒
			//		cos(θ/2) = √(1/3)
			//		sin(θ/2) = √(2/3)
			//
			//	The double-angle formula then gives
			//
			//		cos θ = cos²(θ/2) - sin²(θ/2) = -1/3
			//		sin θ = 2 sin(θ/2) cos(θ/2)   = 2 √2 / 3
			//
			//	For our purposes cos θ and sin θ will be more useful
			//	than the angle θ = 109.47…° itself.
			//
			//	If we draw some O-H-O segments vertically,
			//	then all the rest will be tilted,
			//	with horizontal projection of length theIceOO  sin θ
			//	 and  vertical  projection of length theIceOO |cos θ|

			let theIceHexTileCosineAbs = 1.0/3.0	//	= absolute value of cos(tilt) = -1/3
			let theIceHexTileSine = 2.0 * root2 / 3.0
			let theIceHexTiltedProjH = theIceOO * theIceHexTileSine
			let theIceHexTiltedProjV = theIceOO * theIceHexTileCosineAbs
			let theIceHexTiltedProjX = 0.5 * root3 * theIceHexTiltedProjH
			let theIceHexTiltedProjY =      0.5    * theIceHexTiltedProjH
			let theIceHexAB = 2.0 * theIceHexTiltedProjX
				//	crystallographic translation distances a and b
			let theIceHexC = 2.0 * theIceOO  +  2.0 * theIceHexTiltedProjV
				//	crystallographic translation distance c
			
			//	Note:  The crystallographic translation distances
			//	theIceHexAB and theIceHexC agree to within 1% with the values
			//
			//			a = b = 4.4975Å
			//			  c   = 7.3224Å
			//
			//	found at
			//
			//		https://www1.lsbu.ac.uk/water/ice_crystals.html
			//
			//	And of course because we know theIceOO = 2.76 only to within ~1%,
			//	we wouldn't expect any more precise agreement for a, b and c.
		
			return CrystalData(
				
				formalNameKey: "Ice (hexagonal)",
				familiarNameKey: "Hexagonal ice",
				formula: "H₂O",
				
				atoms: [
				
					Atom(
						element: .oxygen,
						radius: theIceOxygenRadius,
						center: SIMD3<Double>(
							0.0,
							-0.5 * theIceHexTiltedProjH,
							-(0.5 * theIceOO  +  theIceHexTiltedProjV)
						)
					),
				
					Atom(
						element: .oxygen,
						radius: theIceOxygenRadius,
						center: SIMD3<Double>(
							0.0,
							-0.5 * theIceHexTiltedProjH,
							+(0.5 * theIceOO  +  theIceHexTiltedProjV)
						)
					),
				
					Atom(
						element: .oxygen,
						radius: theIceOxygenRadius,
						center: SIMD3<Double>(
							0.0,
							+0.5 * theIceHexTiltedProjH,
							-0.5 * theIceOO
						)
					),
				
					Atom(
						element: .oxygen,
						radius: theIceOxygenRadius,
						center: SIMD3<Double>(
							0.0,
							+0.5 * theIceHexTiltedProjH,
							+0.5 * theIceOO
						)
					),
				
					Atom(
						element: .hydrogen,
						radius: theIceHydrogenRadius,
						center: SIMD3<Double>(
							0.0,
							-0.5 * theIceHexTiltedProjH,
							+(0.5 * theIceOO  +  theIceHexTiltedProjV)
								+ theIceCovalentBondLength
						)
					),
				
					Atom(
						element: .hydrogen,
						radius: theIceHydrogenRadius,
						center: SIMD3<Double>(
							0.0,
							-0.5 * theIceHexTiltedProjH
								+ theIceCovalentBondFraction * theIceHexTiltedProjH,
							+(0.5 * theIceOO  +  theIceHexTiltedProjV)
								- theIceCovalentBondFraction * theIceHexTiltedProjV
						)
					),
				
					Atom(
						element: .hydrogen,
						radius: theIceHydrogenRadius,
						center: SIMD3<Double>(
							0.0,
							+0.5 * theIceHexTiltedProjH,
							-0.5 * theIceOO
								+ theIceCovalentBondLength
						)
					),
				
					Atom(
						element: .hydrogen,
						radius: theIceHydrogenRadius,
						center: SIMD3<Double>(
							0.0,
							+0.5 * theIceHexTiltedProjH
								- theIceCovalentBondFraction * theIceHexTiltedProjH,
							-0.5 * theIceOO
								- theIceCovalentBondFraction * theIceHexTiltedProjV
						)
					),
				
					Atom(
						element: .hydrogen,
						radius: theIceHydrogenRadius,
						center: SIMD3<Double>(
							0.0
								- (1.0 - theIceCovalentBondFraction) * theIceHexTiltedProjX,
							+0.5 * theIceHexTiltedProjH
								+ (1.0 - theIceCovalentBondFraction) * theIceHexTiltedProjY,
							-0.5 * theIceOO
								- (1.0 - theIceCovalentBondFraction) * theIceHexTiltedProjV
						)
					),
				
					Atom(
						element: .hydrogen,
						radius: theIceHydrogenRadius,
						center: SIMD3<Double>(
							0.0
								+ (1.0 - theIceCovalentBondFraction) * theIceHexTiltedProjX,
							+0.5 * theIceHexTiltedProjH
								+ (1.0 - theIceCovalentBondFraction) * theIceHexTiltedProjY,
							-0.5 * theIceOO
								- (1.0 - theIceCovalentBondFraction) * theIceHexTiltedProjV
						)
					),
				
					Atom(
						element: .hydrogen,
						radius: theIceHydrogenRadius,
						center: SIMD3<Double>(
							0.0
								- theIceCovalentBondFraction * theIceHexTiltedProjX,
							+0.5 * theIceHexTiltedProjH
								+ theIceCovalentBondFraction * theIceHexTiltedProjY,
							+0.5 * theIceOO
								+ theIceCovalentBondFraction * theIceHexTiltedProjV
						)
					),
				
					Atom(
						element: .hydrogen,
						radius: theIceHydrogenRadius,
						center: SIMD3<Double>(
							0.0
								+ theIceCovalentBondFraction * theIceHexTiltedProjX,
							+0.5 * theIceHexTiltedProjH
								+ theIceCovalentBondFraction * theIceHexTiltedProjY,
							+0.5 * theIceOO
								+ theIceCovalentBondFraction * theIceHexTiltedProjV
						)
					)
				],
				
				bonds: [
				
					Bond(
						type: .covalent,
						start: SIMD3<Double>(
							0.0,
							-0.5 * theIceHexTiltedProjH,
							+(0.5 * theIceOO  +  theIceHexTiltedProjV)
						),
						finish: SIMD3<Double>(
							0.0,
							-0.5 * theIceHexTiltedProjH,
							+(0.5 * theIceOO  +  theIceHexTiltedProjV)
								+ theIceCovalentBondLength
						)
					),
				
					Bond(
						type: .hydrogen,
						start: SIMD3<Double>(
							0.0,
							-0.5 * theIceHexTiltedProjH,
							+(0.5 * theIceOO  +  theIceHexTiltedProjV)
								+ theIceCovalentBondLength
						),
						finish: SIMD3<Double>(
							0.0,
							-0.5 * theIceHexTiltedProjH,
							+(0.5 * theIceOO  +  theIceHexTiltedProjV)
								+ theIceOO
						)
					),
				
					Bond(
						type: .covalent,
						start: SIMD3<Double>(
							0.0,
							-0.5 * theIceHexTiltedProjH,
							+(0.5 * theIceOO  +  theIceHexTiltedProjV)
						),
						finish: SIMD3<Double>(
							0.0,
							-0.5 * theIceHexTiltedProjH
								+ theIceCovalentBondFraction * theIceHexTiltedProjH,
							+(0.5 * theIceOO  +  theIceHexTiltedProjV)
								- theIceCovalentBondFraction * theIceHexTiltedProjV
						)
					),
				
					Bond(
						type: .hydrogen,
						start: SIMD3<Double>(
							0.0,
							-0.5 * theIceHexTiltedProjH
								+ theIceCovalentBondFraction * theIceHexTiltedProjH,
							+(0.5 * theIceOO  +  theIceHexTiltedProjV)
								- theIceCovalentBondFraction * theIceHexTiltedProjV
						),
						finish: SIMD3<Double>(
							0.0,
							-0.5 * theIceHexTiltedProjH
								+ theIceHexTiltedProjH,
							+(0.5 * theIceOO  +  theIceHexTiltedProjV)
								- theIceHexTiltedProjV
						)
					),
				
					Bond(
						type: .covalent,
						start: SIMD3<Double>(
							0.0,
							+0.5 * theIceHexTiltedProjH,
							-0.5 * theIceOO
						),
						finish: SIMD3<Double>(
							0.0,
							+0.5 * theIceHexTiltedProjH,
							-0.5 * theIceOO
								+ theIceCovalentBondLength
						)
					),
				
					Bond(
						type: .hydrogen,
						start: SIMD3<Double>(
							0.0,
							+0.5 * theIceHexTiltedProjH,
							-0.5 * theIceOO
								+ theIceCovalentBondLength
						),
						finish: SIMD3<Double>(
							0.0,
							+0.5 * theIceHexTiltedProjH,
							-0.5 * theIceOO
								+ theIceOO
						)
					),
				
					Bond(
						type: .covalent,
						start: SIMD3<Double>(
							0.0,
							+0.5 * theIceHexTiltedProjH,
							-0.5 * theIceOO
						),
						finish: SIMD3<Double>(
							0.0,
							+0.5 * theIceHexTiltedProjH
								- theIceCovalentBondFraction * theIceHexTiltedProjH,
							-0.5 * theIceOO
								- theIceCovalentBondFraction * theIceHexTiltedProjV
						)
					),
				
					Bond(
						type: .hydrogen,
						start: SIMD3<Double>(
							0.0,
							+0.5 * theIceHexTiltedProjH
								- theIceCovalentBondFraction * theIceHexTiltedProjH,
							-0.5 * theIceOO
								- theIceCovalentBondFraction * theIceHexTiltedProjV
						),
						finish: SIMD3<Double>(
							0.0,
							+0.5 * theIceHexTiltedProjH
								- theIceHexTiltedProjH,
							-0.5 * theIceOO
								- theIceHexTiltedProjV
						)
					),
				
					Bond(
						type: .hydrogen,
						start: SIMD3<Double>(
							0.0,
							+0.5 * theIceHexTiltedProjH,
							-0.5 * theIceOO
						),
						finish: SIMD3<Double>(
							0.0
								- (1.0 - theIceCovalentBondFraction) * theIceHexTiltedProjX,
							+0.5 * theIceHexTiltedProjH
								+ (1.0 - theIceCovalentBondFraction) * theIceHexTiltedProjY,
							-0.5 * theIceOO
								- (1.0 - theIceCovalentBondFraction) * theIceHexTiltedProjV
						)
					),
				
					Bond(
						type: .covalent,
						start: SIMD3<Double>(
							0.0
								- (1.0 - theIceCovalentBondFraction) * theIceHexTiltedProjX,
							+0.5 * theIceHexTiltedProjH
								+ (1.0 - theIceCovalentBondFraction) * theIceHexTiltedProjY,
							-0.5 * theIceOO
								- (1.0 - theIceCovalentBondFraction) * theIceHexTiltedProjV
						),
						finish: SIMD3<Double>(
							0.0
								- theIceHexTiltedProjX,
							+0.5 * theIceHexTiltedProjH
								+ theIceHexTiltedProjY,
							-0.5 * theIceOO
								- theIceHexTiltedProjV
						)
					),
				
					Bond(
						type: .hydrogen,
						start: SIMD3<Double>(
							0.0,
							+0.5 * theIceHexTiltedProjH,
							-0.5 * theIceOO
						),
						finish: SIMD3<Double>(
							0.0
								+ (1.0 - theIceCovalentBondFraction) * theIceHexTiltedProjX,
							+0.5 * theIceHexTiltedProjH
								+ (1.0 - theIceCovalentBondFraction) * theIceHexTiltedProjY,
							-0.5 * theIceOO
								- (1.0 - theIceCovalentBondFraction) * theIceHexTiltedProjV
						)
					),
				
					Bond(
						type: .covalent,
						start: SIMD3<Double>(
							0.0
								+ (1.0 - theIceCovalentBondFraction) * theIceHexTiltedProjX,
							+0.5 * theIceHexTiltedProjH
								+ (1.0 - theIceCovalentBondFraction) * theIceHexTiltedProjY,
							-0.5 * theIceOO
								- (1.0 - theIceCovalentBondFraction) * theIceHexTiltedProjV
						),
						finish: SIMD3<Double>(
							0.0
								+ theIceHexTiltedProjX,
							+0.5 * theIceHexTiltedProjH
								+ theIceHexTiltedProjY,
							-0.5 * theIceOO
								- theIceHexTiltedProjV
						)
					),
				
					Bond(
						type: .covalent,
						start: SIMD3<Double>(
							0.0,
							+0.5 * theIceHexTiltedProjH,
							+0.5 * theIceOO
						),
						finish: SIMD3<Double>(
							0.0
								- theIceCovalentBondFraction * theIceHexTiltedProjX,
							+0.5 * theIceHexTiltedProjH
								+ theIceCovalentBondFraction * theIceHexTiltedProjY,
							+0.5 * theIceOO
								+ theIceCovalentBondFraction * theIceHexTiltedProjV
						)
					),
				
					Bond(
						type: .hydrogen,
						start: SIMD3<Double>(
							0.0
								- theIceCovalentBondFraction * theIceHexTiltedProjX,
							+0.5 * theIceHexTiltedProjH
								+ theIceCovalentBondFraction * theIceHexTiltedProjY,
							+0.5 * theIceOO
								+ theIceCovalentBondFraction * theIceHexTiltedProjV
						),
						finish: SIMD3<Double>(
							0.0
								- theIceHexTiltedProjX,
							+0.5 * theIceHexTiltedProjH
								+ theIceHexTiltedProjY,
							+0.5 * theIceOO
								+ theIceHexTiltedProjV
						)
					),
				
					Bond(
						type: .covalent,
						start: SIMD3<Double>(
							0.0,
							+0.5 * theIceHexTiltedProjH,
							+0.5 * theIceOO
						),
						finish: SIMD3<Double>(
							0.0
								+ theIceCovalentBondFraction * theIceHexTiltedProjX,
							+0.5 * theIceHexTiltedProjH
								+ theIceCovalentBondFraction * theIceHexTiltedProjY,
							+0.5 * theIceOO
								+ theIceCovalentBondFraction * theIceHexTiltedProjV
						)
					),
				
					Bond(
						type: .hydrogen,
						start: SIMD3<Double>(
							0.0
								+ theIceCovalentBondFraction * theIceHexTiltedProjX,
							+0.5 * theIceHexTiltedProjH
								+ theIceCovalentBondFraction * theIceHexTiltedProjY,
							+0.5 * theIceOO
								+ theIceCovalentBondFraction * theIceHexTiltedProjV
						),
						finish: SIMD3<Double>(
							0.0
								+ theIceHexTiltedProjX,
							+0.5 * theIceHexTiltedProjH
								+ theIceHexTiltedProjY,
							+0.5 * theIceOO
								+ theIceHexTiltedProjV
						)
					),
				],
				
				primitiveTranslations: simd_double3x3(
						SIMD3<Double>(0.0, 0.0, theIceHexC),
						SIMD3<Double>(+0.5 * theIceHexAB, 0.5 * root3 * theIceHexAB, 0.0),
						SIMD3<Double>(-0.5 * theIceHexAB, 0.5 * root3 * theIceHexAB, 0.0)
					),
					//	Remaining face-pairing translation:
					//	SIMD3<Double>(theIceHexAB, 0.0, 0.0)

				aestheticOffset: SIMD3<Double>(theIceHexTiltedProjX, 0.0, 0.0)
			)
				
		case .iceCubic:
		
			//	Cubic ice may be laid out on a rectangular grid.
			let icc = (2.0 * theIceOO / root3) // = length of cubic ice grid cell in ångströms
		
			return CrystalData(
				
				formalNameKey: "Ice (cubic)",
				familiarNameKey: "Cubic ice",
				formula: "H₂O",
				
				atoms: [
				
					Atom(
						element: .oxygen,
						radius: theIceOxygenRadius,
						center: SIMD3<Double>( -0.25*icc, -0.25*icc, -0.25*icc )
					),
				
					Atom(
						element: .oxygen,
						radius: theIceOxygenRadius,
						center: SIMD3<Double>( +0.25*icc, +0.25*icc, +0.25*icc )
					),
				
					Atom(
						element: .hydrogen,
						radius: theIceHydrogenRadius,
						center: SIMD3<Double>(
							-0.25*icc + theIceCovalentBondLength / root3,
							-0.25*icc + theIceCovalentBondLength / root3,
							-0.25*icc + theIceCovalentBondLength / root3
						)
					),
				
					Atom(
						element: .hydrogen,
						radius: theIceHydrogenRadius,
						center: SIMD3<Double>(
							-0.25*icc - theIceCovalentBondLength / root3,
							-0.25*icc - theIceCovalentBondLength / root3,
							-0.25*icc + theIceCovalentBondLength / root3
						)
					),
				
					Atom(
						element: .hydrogen,
						radius: theIceHydrogenRadius,
						center: SIMD3<Double>(
							+0.25*icc - theIceCovalentBondLength / root3,
							+0.25*icc + theIceCovalentBondLength / root3,
							+0.25*icc + theIceCovalentBondLength / root3
						)
					),
				
					Atom(
						element: .hydrogen,
						radius: theIceHydrogenRadius,
						center: SIMD3<Double>(
							+0.25*icc + theIceCovalentBondLength / root3,
							+0.25*icc - theIceCovalentBondLength / root3,
							+0.25*icc + theIceCovalentBondLength / root3
						)
					)
				],
				
				bonds: [
				
					Bond(
						type: .covalent,
						start: SIMD3<Double>( -0.25*icc, -0.25*icc, -0.25*icc ),
						finish: SIMD3<Double>(
							-0.25*icc - theIceCovalentBondLength / root3,
							-0.25*icc - theIceCovalentBondLength / root3,
							-0.25*icc + theIceCovalentBondLength / root3
						)
					),
				
					Bond(
						type: .covalent,
						start: SIMD3<Double>( -0.25*icc, -0.25*icc, -0.25*icc ),
						finish: SIMD3<Double>(
							-0.25*icc + theIceCovalentBondLength / root3,
							-0.25*icc + theIceCovalentBondLength / root3,
							-0.25*icc + theIceCovalentBondLength / root3
						)
					),
				
					Bond(
						type: .hydrogen,
						start: SIMD3<Double>( -0.25*icc, -0.25*icc, -0.25*icc ),
						finish: SIMD3<Double>(
							-0.25*icc - theIceHydrogenBondLength / root3,
							-0.25*icc + theIceHydrogenBondLength / root3,
							-0.25*icc - theIceHydrogenBondLength / root3
						)
					),
				
					Bond(
						type: .hydrogen,
						start: SIMD3<Double>( -0.25*icc, -0.25*icc, -0.25*icc ),
						finish: SIMD3<Double>(
							-0.25*icc + theIceHydrogenBondLength / root3,
							-0.25*icc - theIceHydrogenBondLength / root3,
							-0.25*icc - theIceHydrogenBondLength / root3
						)
					),
				
					Bond(
						type: .covalent,
						start: SIMD3<Double>( +0.25*icc, +0.25*icc, +0.25*icc ),
						finish: SIMD3<Double>(
							+0.25*icc - theIceCovalentBondLength / root3,
							+0.25*icc + theIceCovalentBondLength / root3,
							+0.25*icc + theIceCovalentBondLength / root3
						)
					),
				
					Bond(
						type: .covalent,
						start: SIMD3<Double>( +0.25*icc, +0.25*icc, +0.25*icc ),
						finish: SIMD3<Double>(
							+0.25*icc + theIceCovalentBondLength / root3,
							+0.25*icc - theIceCovalentBondLength / root3,
							+0.25*icc + theIceCovalentBondLength / root3
						)
					),
				
					Bond(
						type: .hydrogen,
						start: SIMD3<Double>( +0.25*icc, +0.25*icc, +0.25*icc ),
						finish: SIMD3<Double>(
							+0.25*icc - theIceHydrogenBondLength / root3,
							+0.25*icc - theIceHydrogenBondLength / root3,
							+0.25*icc - theIceHydrogenBondLength / root3
						)
					),
				
					Bond(
						type: .hydrogen,
						start: SIMD3<Double>( +0.25*icc, +0.25*icc, +0.25*icc ),
						finish: SIMD3<Double>(
							+0.25*icc + theIceHydrogenBondLength / root3,
							+0.25*icc + theIceHydrogenBondLength / root3,
							+0.25*icc - theIceHydrogenBondLength / root3
						)
					)
				],
				
				primitiveTranslations: simd_double3x3(
						SIMD3<Double>( 0.0,  icc,  icc),
						SIMD3<Double>( icc,  0.0,  icc),
						SIMD3<Double>( icc,  icc,  0.0)
					),
					//	Remaining face-pairing translations:
					//	SIMD3<Double>( 0.0,  icc, -icc)
					//	SIMD3<Double>(-icc,  0.0,  icc)
					//	SIMD3<Double>( icc, -icc,  0.0)

				aestheticOffset: SIMD3<Double>( +0.25*icc, +0.25*icc, -0.75*icc )
			)
			
		case .halite:
		
			//	Halite may be laid out on a rectangular grid.
			let hlt = 2.8201 // = 5.6402 Å / 2 = length of halite grid cell in ångströms
		
			//	The raw ionic radii in halite are taken from a table that appears
			//	in some lecture notes online, which in turn attributes the table
			//	to R.D. Shannon, Acta Cryst. A32, 751 (1976).
			//	Elsewhere one finds halite's lattice parameter a = 5.6402,
			//	and that's what I used for the grid cell length hlt = 5.6402 / 2.
			//	So let's tweak the raw radii to exactly fit
			//	the (apparently more precise) lattice parameter.
			let theHaliteSodiumRadiusRaw = 1.02	// radius of Na⁺ ion in halite, in ångströms
			let theHaliteChlorideRadiusRaw = 1.81	// radius of Cl⁻ ion in halite, in ångströms
			let theHaliteTweak = hlt / (theHaliteSodiumRadiusRaw + theHaliteChlorideRadiusRaw)
			let theHaliteSodiumRadius = theAtomEnlargementFactor * theHaliteTweak * theHaliteSodiumRadiusRaw // radius of Na⁺ ion in halite, in ångströms
			let theHaliteChlorideRadius = theAtomEnlargementFactor * theHaliteTweak * theHaliteChlorideRadiusRaw // radius of Cl⁻ ion in halite, in ångströms

			return CrystalData(
				
				formalNameKey: "Halite",
				familiarNameKey: "Salt",
				formula: "NaCl",
				
				atoms: [
				
					Atom(
						element: .chlorine,
						radius: theHaliteChlorideRadius,
						center: SIMD3<Double>( 0.0, 0.0, 0.0 )
					),
				
					Atom(
						element: .sodium,
						radius: theHaliteSodiumRadius,
						center: SIMD3<Double>( 0.0, 0.0, hlt )
					)
				],
				
				bonds: [
				
					Bond(
						type: .ionic,
						start:  SIMD3<Double>(  0.0,  0.0,  0.0 ),
						finish: SIMD3<Double>( -hlt,  0.0,  0.0 )
					),
				
					Bond(
						type: .ionic,
						start:  SIMD3<Double>(  0.0,  0.0,  0.0 ),
						finish: SIMD3<Double>( +hlt,  0.0,  0.0 )
					),
				
					Bond(
						type: .ionic,
						start:  SIMD3<Double>(  0.0,  0.0,  0.0 ),
						finish: SIMD3<Double>(  0.0, -hlt,  0.0 )
					),
				
					Bond(
						type: .ionic,
						start:  SIMD3<Double>(  0.0,  0.0,  0.0 ),
						finish: SIMD3<Double>(  0.0, +hlt,  0.0 )
					),
				
					Bond(
						type: .ionic,
						start:  SIMD3<Double>(  0.0,  0.0,  0.0 ),
						finish: SIMD3<Double>(  0.0,  0.0, -hlt )
					),
				
					Bond(
						type: .ionic,
						start:  SIMD3<Double>(  0.0,  0.0,  0.0 ),
						finish: SIMD3<Double>(  0.0,  0.0, +hlt )
					)
				],
				
				primitiveTranslations: simd_double3x3(
						SIMD3<Double>( 0.0,  hlt,  hlt),
						SIMD3<Double>( hlt,  0.0,  hlt),
						SIMD3<Double>( hlt,  hlt,  0.0)
					),
					//	Remaining face-pairing translations:
					//	SIMD3<Double>( 0.0,  hlt, -hlt)
					//	SIMD3<Double>(-hlt,  0.0,  hlt)
					//	SIMD3<Double>( hlt, -hlt,  0.0)

				//	Shift the user's initial placement hlt/2 ångströms in each direction,
				//	to stand at the center of a cell, looking down a square tube.
				//
				aestheticOffset: SIMD3<Double>(0.5 * hlt, 0.5 * hlt, 0.5 * hlt)
			)
			
		case .fluorite:
		
			//	Fluorite may be laid out on a rectangular grid.
			let flr = 2.7313 // = 5.4626 Å / 2 = length of fluorite grid cell in ångströms
		
			//	Tweak the nominal Ca²⁺ and F⁻ radii to sum exactly
			//	to flr * √3 / 2, so that the ions will be exactly tangent.
			//
			//		Raw radii taken from a source which I can't seem to get
			//		an exact URL for, but which attributes the values to
			//		R. D. Shannon [Acta Crystallogr. A32, 751-767 (1976)].
			//
			let theFluoriteCalciumRadiusRaw = 1.12 // radius of Ca²⁺ ion with coordination number 8 (as in fluorite), in ångströms
			let theFluoriteFluorideRadiusRaw = 1.31 // radius of  F⁻  ion with coordination number 4 (as in fluorite), in ångströms
			let theFluoriteTweak = (flr * root3 * 0.5) / (theFluoriteCalciumRadiusRaw + theFluoriteFluorideRadiusRaw)
			let theFluoriteCalciumRadius = theAtomEnlargementFactor * theFluoriteTweak * theFluoriteCalciumRadiusRaw // radius of Ca²⁺ ion in fluorite, in ångströms
			let theFluoriteFluorideRadius = theAtomEnlargementFactor * theFluoriteTweak * theFluoriteFluorideRadiusRaw // radius of  F⁻  ion in fluorite, in ångströms

			return CrystalData(
				
				formalNameKey: "Fluorite",
				familiarNameKey: "Fluorite",
				formula: "CaF₂",
				
				atoms: [
				
					Atom(
						element: .calcium,	//	Ca²⁺
						radius: theFluoriteCalciumRadius,
						center: SIMD3<Double>(  0.0*flr,  0.0*flr,  0.0*flr )
					),
				
					Atom(
						element: .fluorine,	//	F⁻
						radius: theFluoriteFluorideRadius,
						center: SIMD3<Double>( +0.5*flr, +0.5*flr, +0.5*flr )
					),
				
					Atom(
						element: .fluorine,	//	F⁻
						radius: theFluoriteFluorideRadius,
						center: SIMD3<Double>( -0.5*flr, -0.5*flr, -0.5*flr )
					)
				],
				
				bonds: [
				
					Bond(
						type: .ionic,
						start:  SIMD3<Double>(  0.0*flr,  0.0*flr,  0.0*flr ),
						finish: SIMD3<Double>( -0.5*flr, -0.5*flr, -0.5*flr )
					),
				
					Bond(
						type: .ionic,
						start:  SIMD3<Double>(  0.0*flr,  0.0*flr,  0.0*flr ),
						finish: SIMD3<Double>( +0.5*flr, -0.5*flr, -0.5*flr )
					),

					Bond(
						type: .ionic,
						start:  SIMD3<Double>(  0.0*flr,  0.0*flr,  0.0*flr ),
						finish: SIMD3<Double>( -0.5*flr, +0.5*flr, -0.5*flr )
					),
					
					Bond(
						type: .ionic,
						start:  SIMD3<Double>(  0.0*flr,  0.0*flr,  0.0*flr ),
						finish: SIMD3<Double>( -0.5*flr, -0.5*flr, +0.5*flr )
					),

					Bond(
						type: .ionic,
						start:  SIMD3<Double>(  0.0*flr,  0.0*flr,  0.0*flr ),
						finish: SIMD3<Double>( -0.5*flr, +0.5*flr, +0.5*flr )
					),

					Bond(
						type: .ionic,
						start:  SIMD3<Double>(  0.0*flr,  0.0*flr,  0.0*flr ),
						finish: SIMD3<Double>( +0.5*flr, -0.5*flr, +0.5*flr )
					),
				
					Bond(
						type: .ionic,
						start:  SIMD3<Double>(  0.0*flr,  0.0*flr,  0.0*flr ),
						finish: SIMD3<Double>( +0.5*flr, +0.5*flr, -0.5*flr )
					),
				
					Bond(
						type: .ionic,
						start:  SIMD3<Double>(  0.0*flr,  0.0*flr,  0.0*flr ),
						finish: SIMD3<Double>( +0.5*flr, +0.5*flr, +0.5*flr )
					)
				],
				
				primitiveTranslations: simd_double3x3(
						SIMD3<Double>( 0.0,  flr,  flr),
						SIMD3<Double>( flr,  0.0,  flr),
						SIMD3<Double>( flr,  flr,  0.0)
					),
					//	Remaining face-pairing translations:
					//	SIMD3<Double>( 0.0,  flr, -flr)
					//	SIMD3<Double>(-flr,  0.0,  flr)
					//	SIMD3<Double>( flr, -flr,  0.0)

				aestheticOffset: SIMD3<Double>(0.0, flr, 0.0)
			)
			
		case .diamond:
		
			//	The diamond crystal may be laid out on a rectangular grid.
			//	Let's define a constant giving the length of the grid cell
			//	in ångströms, so that we may then express the various
			//	atom and bond placements as simple fractions of that grid length.
			let dmd = 1.784 // = 1.545  Å / (0.5 √3)
							// = length of diamond grid cell in ångströms

			//	The web page
			//
			//		https://hypertextbook.com/facts/2001/AliceWarrenGregory.shtml
			//
			//	cites
			//
			//		Brown, LeMay & Bursten. Chemistry The Central Science.
			//		Upper Saddle River, NJ: Simon & Schuster, 1997: 227, 412-413.
			//
			//	as saying
			//
			//		“Atomic radii allow us to estimate the bond lengths
			//		between different elements in molecules. For example,
			//		the C-C bond length in elemental carbon (diamond) is 1.54 Å,
			//		implying a radius of 0.77 Å.”
			//
			//	(although it would seem to be the other way around,
			//	that the bond lengths allow us to estimate the atomic radii).
			//	That same page cites several other sources giving
			//	almost the same result, for example
			//
			//		Saada, David.  https://phycomp.technion.ac.il/~david/thesis/node3.html
			//		22 June 2000.
			//
			//	gives the more precise values
			//
			//		                         diamond    graphite
			//		Lattice constant (RT)    3.567 Å    2.462 Å, 6.708 Å (the latter being twice the interlayer spacing in AB graphite)
			//		Bond length (RT)         1.545 Å    1.421 Å
			//
			let theDiamondCarbonRadius = theAtomEnlargementFactor * 0.25 * root3 * dmd

			return CrystalData(
				
				formalNameKey: "Diamond",
				familiarNameKey: "Diamond",
				formula: "C",
				
				atoms: [
				
					Atom(
						element: .carbon,
						radius: theDiamondCarbonRadius,
						center: SIMD3<Double>( -0.25*dmd, -0.25*dmd, -0.25*dmd )
					),
				
					Atom(
						element: .carbon,
						radius: theDiamondCarbonRadius,
						center: SIMD3<Double>( +0.25*dmd, +0.25*dmd, +0.25*dmd )
					)
				],
				
				bonds: [
				
					Bond(
						type: .covalent,
						start:  SIMD3<Double>( -0.25*dmd, -0.25*dmd, -0.25*dmd ),
						finish: SIMD3<Double>( +0.25*dmd, +0.25*dmd, +0.25*dmd )
					),
				
					Bond(
						type: .covalent,
						start:  SIMD3<Double>( -0.25*dmd, -0.25*dmd, -0.25*dmd ),
						finish: SIMD3<Double>( +0.25*dmd, -0.75*dmd, -0.75*dmd )
					),
				
					Bond(
						type: .covalent,
						start:  SIMD3<Double>( -0.25*dmd, -0.25*dmd, -0.25*dmd ),
						finish: SIMD3<Double>( -0.75*dmd, +0.25*dmd, -0.75*dmd )
					),
				
					Bond(
						type: .covalent,
						start:  SIMD3<Double>( -0.25*dmd, -0.25*dmd, -0.25*dmd ),
						finish: SIMD3<Double>( -0.75*dmd, -0.75*dmd, +0.25*dmd )
					)
				],
				
				primitiveTranslations: simd_double3x3(
						SIMD3<Double>( 0.0,  dmd,  dmd),
						SIMD3<Double>( dmd,  0.0,  dmd),
						SIMD3<Double>( dmd,  dmd,  0.0)
					),
					//	Remaining face-pairing translations:
					//	SIMD3<Double>( 0.0,  dmd, -dmd)
					//	SIMD3<Double>(-dmd,  0.0,  dmd)
					//	SIMD3<Double>( dmd, -dmd,  0.0)

				//	Place the user's initial head position at the point
				//	where two "orthogonally skew" tunnels overlap each other.
				aestheticOffset: SIMD3<Double>( +0.25*dmd, 0.0, -0.75*dmd )
			)
			
		case .quartz:
		
			//	Quartz may be laid out on a hexagonal grid.
			let qtza = 4.913	//	= length of quartz "horizontal" hexagonal
			//					//	  translation (a₁ = a₂ = a₃) in ångströms
		//	let qtzc = 5.405	//	= length of quartz "vertical" translation (c)
								//	  in ångströms

			let theQuartzCosθ = 0.959930	//	corresponds to θ = 0.284040 for
			let theQuartzSinθ = 0.280236	//		bond angle 144° at oxygen
		//	let theQuartzR = theQuartzCosθ + 1.0/root3
		//	let theQuartzS = (1.0/root3) * theQuartzSinθ
		//	let theQuartzH = 2.0 * theQuartzCosθ

			//	Regarding the ratio
			//
			//		 vertical  translation distance c
			//		---------------------------------
			//		horizontal translation distance a
			//
			//	the geometrically calculated ratio
			//
			//			   3 cos θ
			//			------------ = 1.0815
			//			√3 cos θ + 1
			//
			//	where θ is the bond angle at the Si-O-Si junction,
			//	agrees sufficiently well with the measured ratio
			//
			//			5.405Å
			//			------ = 1.1001
			//			4.913Å
			//
			//	(Somewhere I might have some notes (on paper!)
			//	about how I calculated the geometrical version,
			//	but they're not on my desk as I write this.)
			//
			//		Note:  The measured ratio may include experimental error
			//		as well as any deviations from pure tetrahedral symmetry
			//		at the Si atoms.  So we can't use the measured ratio
			//		to deduce a value for θ.  (I tried, and got cos θ > 1.)
			//
			//	Let's go ahead and use the geometrically calculated ratio,
			//	to ensure consistency with the computed positions for the atoms.
			//
			let gQuartzAspectRatio = 3.0 * theQuartzCosθ / (root3*theQuartzCosθ + 1.0)

			//	The page
			//
			//		https://en.wikipedia.org/wiki/Silicon_dioxide
			//
			//	says
			//
			//		“in α-quartz the bond length is 161 pm”
			//
			//	while our geometrically computed bond length
			//
			//		qtza / (2*(cos(θ) + 1/√3))
			//
			//	comes out to 1.597… .
			//	Either way, some guesswork is required to choose atomic radii
			//	that sum to the bond length, given that "atomic radius" isn't
			//	a well-defined concept to begin with.  The page
			//
			//		https://en.wikipedia.org/wiki/Covalent_radius
			//
			//	gives a covalent radius of 1.11 Å for silicon and 0.66 Å for oxygen,
			//	so let's use those proportions, but with the values rescaled
			//	to sum to 1.597… .
			//
			let theQuartzBondLength = qtza * 0.5 / (theQuartzCosθ + (1.0 / root3))
			let theQuartzSiliconRadiusRaw = 1.11 // in ångströms
			let theQuartzOxygenRadiusRaw = 0.66 // in ångströms
			let theQuartzTweak = theQuartzBondLength / (theQuartzSiliconRadiusRaw + theQuartzOxygenRadiusRaw)
			let theQuartzSiliconRadius = theAtomEnlargementFactor * theQuartzTweak * theQuartzSiliconRadiusRaw // in ångströms
			let theQuartzOxygenRadius = theAtomEnlargementFactor * theQuartzTweak * theQuartzOxygenRadiusRaw // in ångströms

			return CrystalData(
				
				formalNameKey: "Quartz",
				familiarNameKey: "Quartz",
				formula: "SiO₂",
				
				atoms: [
				
					Atom(
						element: .silicon,
						radius: theQuartzSiliconRadius,
						center: SIMD3<Double>(
						
							qtza * (theQuartzCosθ + (1.0/root3))
								 / (2.0*root3*theQuartzCosθ + 2.0),
								 
							qtza * ((1.0/root3) * theQuartzSinθ)
								 / (2.0*root3*theQuartzCosθ + 2.0),
								 
							0.00
						)
					),
				
					Atom(
						element: .oxygen,
						radius: theQuartzOxygenRadius,
						center: SIMD3<Double>(
						
							qtza * ((1.0/root3) - theQuartzSinθ)
								 / (2.0*root3*theQuartzCosθ + 2.0),
								 
							qtza * ((1.0/root3)*theQuartzSinθ + 1.0)
								 / (2.0*root3*theQuartzCosθ + 2.0),
								 
							qtza * (theQuartzCosθ - theQuartzSinθ)
								 / (2.0*root3*theQuartzCosθ + 2.0)
						)
					),
				
					Atom(
						element: .oxygen,
						radius: theQuartzOxygenRadius,
						center: SIMD3<Double>(
						
							qtza * ((1.0/root3) - theQuartzSinθ + 2.0*theQuartzCosθ)
								 / (2.0*root3*theQuartzCosθ + 2.0),
								 
							qtza * ((1.0/root3)*theQuartzSinθ - 1.0)
								 / (2.0*root3*theQuartzCosθ + 2.0),
								 
							qtza * (theQuartzCosθ + theQuartzSinθ)
								 / (2.0*root3*theQuartzCosθ + 2.0)
						)
					)
				],
				
				bonds: [
				
					Bond(
						type: .covalent,
						start: SIMD3<Double>(
						
							qtza * (theQuartzCosθ + (1.0/root3))
								 / (2.0*root3*theQuartzCosθ + 2.0),
								 
							qtza * ((1.0/root3) * theQuartzSinθ)
								 / (2.0*root3*theQuartzCosθ + 2.0),
								 
							0.00
						),
						finish: SIMD3<Double>(
						
							qtza * ((1.0/root3) - theQuartzSinθ)
								 / (2.0*root3*theQuartzCosθ + 2.0),
								 
							qtza * ((1.0/root3)*theQuartzSinθ + 1.0)
								 / (2.0*root3*theQuartzCosθ + 2.0),
								 
							qtza * (theQuartzCosθ - theQuartzSinθ)
								 / (2.0*root3*theQuartzCosθ + 2.0)
						)
					),
				
					Bond(
						type: .covalent,
						start: SIMD3<Double>(
						
							qtza * (theQuartzCosθ + (1.0/root3))
								 / (2.0*root3*theQuartzCosθ + 2.0),
							
							qtza * ((1.0/root3) * theQuartzSinθ)
								 / (2.0*root3*theQuartzCosθ + 2.0),
							
							0.00
						),
						finish: SIMD3<Double>(
						
							qtza * ((1.0/root3) - theQuartzSinθ + 2.0*theQuartzCosθ)
								 / (2.0*root3*theQuartzCosθ + 2.0),
							
							qtza * ((1.0/root3)*theQuartzSinθ - 1.0)
								 / (2.0*root3*theQuartzCosθ + 2.0),
							
							qtza * (theQuartzCosθ + theQuartzSinθ)
								 / (2.0*root3*theQuartzCosθ + 2.0)
						)
					),
				
					Bond(
						type: .covalent,
						start: SIMD3<Double>(
						
							qtza * (theQuartzCosθ + (1.0/root3))
								 / (2.0*root3*theQuartzCosθ + 2.0),
							
							qtza * ((1.0/root3) * theQuartzSinθ)
								 / (2.0*root3*theQuartzCosθ + 2.0),
							
							0.00
						),
						finish: SIMD3<Double>(
						
							qtza * ((1.0/root3) + theQuartzSinθ)
								 / (2.0*root3*theQuartzCosθ + 2.0),
							
							qtza * ((1.0/root3)*theQuartzSinθ - 1.0)
								 / (2.0*root3*theQuartzCosθ + 2.0),
							
							qtza * (-theQuartzCosθ - theQuartzSinθ)
								 / (2.0*root3*theQuartzCosθ + 2.0)
						)
					),
				
					Bond(
						type: .covalent,
						start: SIMD3<Double>(
						
							qtza * (theQuartzCosθ + (1.0/root3))
								 / (2.0*root3*theQuartzCosθ + 2.0),
							
							qtza * ((1.0/root3) * theQuartzSinθ)
								 / (2.0*root3*theQuartzCosθ + 2.0),
							
							0.00
						),
						finish: SIMD3<Double>(
						
							qtza * ((1.0/root3) + theQuartzSinθ + 2.0*theQuartzCosθ)
								 / (2.0*root3*theQuartzCosθ + 2.0),
							
							qtza * ((1.0/root3)*theQuartzSinθ + 1.0)
								 / (2.0*root3*theQuartzCosθ + 2.0),
							
							qtza * (-theQuartzCosθ + theQuartzSinθ)
								 / (2.0*root3*theQuartzCosθ + 2.0)
						)
					)
				],
				
				primitiveTranslations: simd_double3x3(
					//	See the comment "Special treatment for quartz" near the top of this file.
						SIMD3<Double>(0.0, 0.0, gQuartzAspectRatio * qtza),	//	≈ (0, 0, qtzc)
						SIMD3<Double>(+0.5 * root3 * qtza, 0.5 * qtza, 0.0),
						SIMD3<Double>(-0.5 * root3 * qtza, 0.5 * qtza, 0.0)
					),
					//	Remaining face-pairing translation:
					//	SIMD3<Double>(0.0, qtza, 0.0)
				
				corkscrew:
					//	See the comment "Special treatment for quartz" near the top of this file.
					simd_double4x4(
						SIMD4<Double>(   -0.5,     0.5*root3,  0.0,  0.0 ),
						SIMD4<Double>(-0.5*root3,    -0.5,     0.0,  0.0 ),
						SIMD4<Double>(    0.0,        0.0,     1.0,  0.0 ),
						SIMD4<Double>(    0.0,        0.0, (gQuartzAspectRatio * qtza)/3.0, 1.0)	//	≈ (0, 0, qtzc/3, 1)
					),

				aestheticOffset: SIMD3<Double>.zero
			)
		}
	}

	private func generateQuartzAtoms(
	) -> [Atom] {
	
		let theQuartzCrystalData: CrystalData = unifiedData
		
		guard let theCorkscrewMotion = theQuartzCrystalData.corkscrew else {
			return theQuartzCrystalData.atoms	//	should never occur
		}
		
		let theTransformations = [
			theCorkscrewMotion.inverse,
			matrix_identity_double4x4,
			theCorkscrewMotion
		]
		
		var theUnitCellQuartzAtoms: [Atom] = []
		theUnitCellQuartzAtoms.reserveCapacity(theTransformations.count * theQuartzCrystalData.atoms.count)
		
		for theTransformation in theTransformations {
		
			for theAtom in theQuartzCrystalData.atoms {
			
				let theTransformedCenterExt = theTransformation * SIMD4<Double>(theAtom.center, 1.0)
				let theTransformedCenter = SIMD3<Double>(
											theTransformedCenterExt[0],
											theTransformedCenterExt[1],
											theTransformedCenterExt[2])
				
				let theTransformedAtom = Atom(
					element: theAtom.element,
					radius: theAtom.radius,
					center: theTransformedCenter)
				
				theUnitCellQuartzAtoms.append(theTransformedAtom)
			}
		}
		
		return theUnitCellQuartzAtoms
	}

	private func generateQuartzBonds(
	) -> [Bond] {
	
		let theQuartzCrystalData: CrystalData = unifiedData
		
		guard let theCorkscrewMotion = theQuartzCrystalData.corkscrew else {
			return theQuartzCrystalData.bonds	//	should never occur
		}
		
		let theTransformations = [
			theCorkscrewMotion.inverse,
			matrix_identity_double4x4,
			theCorkscrewMotion
		]
		
		var theUnitCellQuartzBonds: [Bond] = []
		theUnitCellQuartzBonds.reserveCapacity(theTransformations.count * theQuartzCrystalData.bonds.count)
		
		for theTransformation in theTransformations {
		
			for theBond in theQuartzCrystalData.bonds {
			
				let theTransformedStartExt = theTransformation * SIMD4<Double>(theBond.start, 1.0)
				let theTransformedStart = SIMD3<Double>(
											theTransformedStartExt[0],
											theTransformedStartExt[1],
											theTransformedStartExt[2])
			
				let theTransformedFinishExt = theTransformation * SIMD4<Double>(theBond.finish, 1.0)
				let theTransformedFinish = SIMD3<Double>(
											theTransformedFinishExt[0],
											theTransformedFinishExt[1],
											theTransformedFinishExt[2])
				
				let theTransformedBond = Bond(
					type: theBond.type,
					start: theTransformedStart,
					finish: theTransformedFinish)
				
				theUnitCellQuartzBonds.append(theTransformedBond)
			}
		}
		
		return theUnitCellQuartzBonds
	}
}


// MARK: -
// MARK: gamma encoding

func gammaEncode(
	_ linearValue: Double
) -> Double {	//	returns gamma-encoded value (using sRGB transfer function)

	//	The encoding curve for negative values is
	//	symmetrical with the curve for positive values.
	//
	if linearValue < 0.0 {
		return -gammaEncode( -linearValue )
	}
	
	//	The sRGB transfer function begins with a short linear segment
	//	(to avoid an infinite slope near zero), followed by a gamma curve
	//	with power 1/2.4 thereafter.  The overall curve closely follows
	//	a traditional gamma curve (with no linear segment) with power 1/2.2.
	//
	//	My understanding is that the following numerical values are exact
	//	(defined in the IEC standard), they're not approximations.
	//
	let theGammaEncodedValue = (
		linearValue < 0.0031308 ?
			12.92 * linearValue :
			1.055 * pow(linearValue, 1.0/2.4) - 0.055
	)

	return theGammaEncodedValue
}

//	Crystal doesn't need this gammaDecode() function, but it
//	seems like a good idea for gammaEncode() and gammaDecode()
//	to travel together, for possible future use in other contexts.
//
func gammaDecode(
	_ gammaEncodedValue: Double	//	encoded using sRGB transfer function
) -> Double {	//	returns linear value

	//	The decoding curve for negative values is
	//	symmetrical with the curve for positive values.
	//
	if gammaEncodedValue < 0.0 {
		return -gammaDecode( -gammaEncodedValue )
	}

	//	The sRGB transfer function begins with a short linear segment
	//	(to avoid zero slope near zero), followed by a gamma curve
	//	with power 2.4 thereafter.  The overall curve closely follows
	//	a traditional gamma curve (with no linear segment) with power 2.2.
	//
	//	My understanding is that the following numerical values are exact
	//	(defined in the IEC standard), they're not approximations.
	//
	let theLinearValue = (
		gammaEncodedValue < 0.04045 ?
			(1.0 / 12.92) * gammaEncodedValue :
			pow((1.0/1.055) * gammaEncodedValue + (0.055/1.055), 2.4)
	)

	return theLinearValue
}
