//	CrystalImmersiveView.swift
//
//	© 2024 by Jeff Weeks
//	See TermsOfUse.txt

import SwiftUI
import RealityKit

struct CrystalImmersiveView: View {

	let crystal: Crystal
	
	//	The tilingRadiusÅ could be an adjustable parameter if desired,
	//	but for now the best strategy might be to just set it to
	//	the largest radius that the Vision Pro can tile out to
	//	while reliably making its frame rate and not overheating.
	let tilingRadiusÅ: Double = 20.0	//	in ångstroms

	let scaleFactor: Double	//	dimensionless, ∈  10⁹ … 10¹⁰
	let showBonds: Bool

	@Binding var atomicRadiusFactor: Double	//	dimensionless, ∈  25% … 100%

	@Environment(\.scenePhase) var scenePhase
	@Environment(\.dismissWindow) var dismissWindow
	@Environment(\.openWindow) var openWindow
	

	var body: some View {
	
		RealityView() { content in

			let theParentEntity = Entity()
			theParentEntity.name = "parent"
			theParentEntity.components[ManifestComponent.self]
				= ManifestComponent(atomicRadiusFactor: atomicRadiusFactor)
			
			let theAtomsEntity = Entity()
			theAtomsEntity.name = "atoms"
			let theAtomEntities = crystal.atoms.map { theAtom in makeAtomEntity(atom: theAtom) }
			theAtomsEntity.children.append(contentsOf: theAtomEntities)
			theParentEntity.addChild(theAtomsEntity)
			
			if showBonds {
				let theBondsEntity = Entity()
				theBondsEntity.name = "bonds"
				let theBondEntities = crystal.bonds.map { theBond in makeBondEntity(bond: theBond) }
				theBondsEntity.children.append(contentsOf: theBondEntities)
				theParentEntity.addChild(theBondsEntity)
			}
			
			let theTapSphereEntity = Entity()
			theTapSphereEntity.name = "tap sphere"
			theTapSphereEntity.components[InputTargetComponent.self]
				= InputTargetComponent(allowedInputTypes: .all)
			theTapSphereEntity.components[CollisionComponent.self]
				= CollisionComponent(
					shapes: [ ShapeResource.generateSphere(radius: 5.0) ],
					filter: CollisionFilter(group: [], mask: [] ) )
			theParentEntity.addChild(theTapSphereEntity)
			
			content.entities.replaceAll([theParentEntity])

		} update: { content in

			guard let theParentEntity = content.entities.first,
			      theParentEntity.name == "parent",
			      let theManifest = theParentEntity.components[ManifestComponent.self]
			else {
				preconditionFailure("Couldn't find theParentEntity and/or theManifest in update:")
			}
			
			if atomicRadiusFactor != theManifest.atomicRadiusFactor {
			
				theParentEntity.components[ManifestComponent.self]
					= ManifestComponent(atomicRadiusFactor: atomicRadiusFactor)
				
				let theAtomEntities = crystal.atoms.map {
					theAtom in makeAtomEntity(atom: theAtom)
				}

				guard let theAtomsEntity = theParentEntity.findEntity(named: "atoms") else {
					preconditionFailure("Couldn't find theAtomsEntity in update:")
				}
				theAtomsEntity.children.replaceAll(theAtomEntities)
			}
		}
		.gesture(doubleTapGesture)
		.gesture(atomMagnifyGesture)
		.onChange(of: scenePhase, initial: true
			//	With "initial: true", when this immersive view first appears
			//	we'll get a 'change' from .active to .active
			//	so we can dismiss the Control Panel.
		) { (oldPhase, newPhase) in
			switch newPhase {
			case .active:	dismissWindow(id: "Control panel", value: 0)
			default:		openWindow(id: "Control panel", value: 0)
			}
		}
	}


// MARK: -
// MARK: Atoms

	func makeAtomEntity(
		atom: Atom
	) -> Entity {
				
		let theAtomModels = makeAtomModels(atom: atom)
		let theAtomInstances = makeVisibleInstances(entityCenter: atom.center)
		
		var theMeshContents = MeshResource.Contents()
		theMeshContents.models = MeshModelCollection(theAtomModels)
		theMeshContents.instances = MeshInstanceCollection(theAtomInstances)

		//	According to the WWDC 2021 talk #10075
		//
		//		“Explore advanced rendering with RealityKit 2”
		//		https://developer.apple.com/videos/play/wwdc2021/10075/
		//
		//	(timepoint 25:12) the following call to MeshResource.generate(from:)
		//	invokes RealityKit's mesh processor (also called a "mesh optimizer"
		//	at timepoint 27:44) to optimize the mesh.  Or maybe
		//		MeshResource.generate(from: theMeshContents)
		//	doesn't invoke the mesh optimizer?  But maybe
		//		MeshResource.Model(id: …, descriptors: [theMeshDescriptor])
		//	does call it?
		//	Apple's official documentation pages don't say, not even for
		//		MeshResource.generate(from: [theMeshDescriptor])
		//	which according to the WWDC video does indeed call the optimizer.
		//
		guard let theAtomMeshResource = try? MeshResource.generate(from: theMeshContents) else {
			preconditionFailure("Couldn't generate theAtomMeshResource")
		}

		let theMaterials = makeMaterials(
			linearDisplayP3Color: atom.element.linearDisplayP3Color,
			roughness: 0.125,
			metallic: 0.0)

		let theAtomModelComponent = ModelComponent(
										mesh: theAtomMeshResource,
										materials: theMaterials)

		let theAtomEntity = Entity()
		theAtomEntity.components[ModelComponent.self] = theAtomModelComponent
		theAtomEntity.position = .zero
		
		return theAtomEntity
	}

	func makeAtomModels(
		atom: Atom
	) -> [MeshResource.Model] {
	
		let theAtomRadiusInÅngströms = atom.radius
		let theAtomRadiusInMeters = 1.0e-10 * theAtomRadiusInÅngströms
		let theSphereRadiusInMeters = scaleFactor * theAtomRadiusInMeters
		let theShrunkSphereRadiusInMeters = atomicRadiusFactor * theSphereRadiusInMeters

		let theUnitSphere = makeUnitSphere(
								baseShape: .icosahedron,
								refinementLevel: 3)
		
		//	For the vertex positions we must scale the unit sphere's vertices
		//	by a factor of theShrunkSphereRadiusInMeters...
		let thePositions = theUnitSphere.positions.map(){ p in
			SIMD3<Float>(p * theShrunkSphereRadiusInMeters)
		}

		//	...but for the normal vectors we may copy the unit sphere's vertices directly.
		let theNormals = theUnitSphere.positions.map(){ p in
			SIMD3<Float>(p)
		}
		
		let theFacets = theUnitSphere.facets

		let theModels = makeModels(
							positions: thePositions,
							normals: theNormals,
							facets: theFacets)
		
		return theModels
	}
	

// MARK: -
// MARK: Bonds

	func makeBondEntity(
		bond: Bond
	) -> Entity {
	
		let theBondModels = makeBondModels(bond: bond)
		let theBondInstances = makeVisibleInstances(
								entityCenter: 0.5 * (bond.start + bond.finish))
		
		var theMeshContents = MeshResource.Contents()
		theMeshContents.models = MeshModelCollection(theBondModels)
		theMeshContents.instances = MeshInstanceCollection(theBondInstances)

		//	According to the WWDC 2021 talk #10075
		//
		//		“Explore advanced rendering with RealityKit 2”
		//		https://developer.apple.com/videos/play/wwdc2021/10075/
		//
		//	(timepoint 25:12) the following call to MeshResource.generate(from:)
		//	invokes RealityKit's mesh processor (also called a "mesh optimizer"
		//	at timepoint 27:44) to optimize the mesh.  Or maybe
		//		MeshResource.generate(from: theMeshContents)
		//	doesn't invoke the mesh optimizer?  But maybe
		//		MeshResource.Model(id: …, descriptors: [theMeshDescriptor])
		//	does call it?
		//	Apple's official documentation pages don't say, not even for
		//		MeshResource.generate(from: [theMeshDescriptor])
		//	which according to the WWDC video does indeed call the optimizer.
		//
		guard let theBondMeshResource = try? MeshResource.generate(from: theMeshContents) else {
			preconditionFailure("Couldn't generate theBondMeshResource")
		}

		let theMaterials = makeMaterials(
			linearDisplayP3Color: bond.type.linearDisplayP3Color,
			roughness: 0.5,
			metallic: 1.0)

		let theBondModelComponent = ModelComponent(
										mesh: theBondMeshResource,
										materials: theMaterials)

		let theBondEntity = Entity()
		theBondEntity.components[ModelComponent.self] = theBondModelComponent
		theBondEntity.position = .zero
		
		return theBondEntity
	}

	func makeBondModels(
		bond: Bond
	) -> [MeshResource.Model] {
	
		let theBondRadiusInÅngströms = bond.type.radius
		let theBondLengthInÅngströms = length(bond.finish - bond.start)
		let theBondRadiusInMeters = 1.0e-10 * theBondRadiusInÅngströms
		let theBondLengthInMeters = 1.0e-10 * theBondLengthInÅngströms
		let theCylinderRadiusInMeters = scaleFactor * theBondRadiusInMeters
		let theCylinderLengthInMeters = scaleFactor * theBondLengthInMeters
		let theCylinderHalfLengthInMeters = 0.5 * theCylinderLengthInMeters
		
		//	theUnitCylinder has radius 1.0, with ends at z = ±1.0
		//	(so its total length is 2.0).
		let theUnitCylinder = makeUnitCylinder(
								style: .antiprism,
								numVerticesPerEnd: 32)

		let theCylinderDirection = SIMD3<Double>(0.0, 0.0, 1.0)
		let theBondDirection = normalize(bond.finish - bond.start)
		let theRotation = simd_quatd(from: theCylinderDirection, to: theBondDirection)

		//	For the vertex positions we must re-scale the unit cylinder's vertices
		//	and apply the required rotation.
		let thePositions = theUnitCylinder.positions.map(){ p in
		
			let theScaledPosition = SIMD3<Double>(
				p[0] * theCylinderRadiusInMeters,
				p[1] * theCylinderRadiusInMeters,
				p[2] * theCylinderHalfLengthInMeters)
				
			let theRotatedPosition = theRotation.act(theScaledPosition)
			
			return SIMD3<Float>(theRotatedPosition)
		}

		//	For the normal vectors, we may take each point (x,y,z)
		//	on the unit cylinder, suppress the z coordinate
		//	to get a unit-length normal vector (x,y,0),
		//	and then rotate it as required.
		let theNormals = theUnitCylinder.positions.map(){ p in
		
			let theRawNormal = SIMD3<Double>(p[0], p[1], 0.0)
			let theRotatedNormal = theRotation.act(theRawNormal)

			return SIMD3<Float>(theRotatedNormal)
		}
		
		let theFacets = theUnitCylinder.facets

		let theModels = makeModels(
							positions: thePositions,
							normals: theNormals,
							facets: theFacets)

		return theModels
	}


// MARK: -
// MARK: Atoms & bonds

	func makeModels(
		positions: [SIMD3<Float>],
		normals: [SIMD3<Float>],
		facets: [UInt32]
	) -> [MeshResource.Model] {
	
		//	The four models differ only in their material index.
		//	The four materials will differ only in their dimness,
		//	to accommodate "faux fog".
		let theModels = ( 0 ..< 4 ).map() { i in
		
			var theMeshDescriptor = MeshDescriptor()
			theMeshDescriptor.name = "single mesh part"
			theMeshDescriptor.positions = MeshBuffer(positions)
			theMeshDescriptor.normals = MeshBuffer(normals)
			theMeshDescriptor.primitives = .triangles(facets)
			theMeshDescriptor.materials = .allFaces(UInt32(i))

			guard let theModel = try? MeshResource.Model(
										id: "model \(i)",
										descriptors: [theMeshDescriptor]) else {
				preconditionFailure("Failed to generate MeshResource.Model")
			}
			
			return theModel
		}
		
		return theModels
	}

	func makeVisibleInstances(
		entityCenter: SIMD3<Double>	//	in ångströms
	) -> [MeshResource.Instance] {

		//	Our task here is to create an Array containing
		//	all the repeating images of the given atom or bond
		//	whose center lies within the tiling sphere,
		//	while excluding all the images whose center doesn't.

		//	We'll work in ångstroms, then convert to meters at the end.

		//	It's convenient to work with two conceptually different spaces.
		//
		//		Physical space is the real world, measured in ångströms,
		//			with a focus on the tiling sphere's interior.
		//
		//		Cubic space is an abstract Euclidean space containing
		//			a unit cubic lattice with one copy of the given atom or bond
		//			at each point with integer coordinates.
		//
		//	The crystal's lattice mapping maps the cubic space's integer points
		//	to the desired locations for the atom or bond in physical space.
		let pt = crystal.primitiveTranslations
		let ic = (entityCenter + crystal.aestheticOffset)	//	instance center
		let theLatticeMapping = simd_double4x4(
			SIMD4<Double>( pt[0][0], pt[0][1], pt[0][2], 0.0),
			SIMD4<Double>( pt[1][0], pt[1][1], pt[1][2], 0.0),
			SIMD4<Double>( pt[2][0], pt[2][1], pt[2][2], 0.0),
			SIMD4<Double>(  ic[0],    ic[1],    ic[2],   1.0)
		)

		//	Our plan is to map the tiling sphere from physical space
		//	back to cubic space, find the minimal and maximal
		//	x, y and z coordinates that it covers, and then
		//	iterate over a grid in cubic space with those
		//	minimal and maximal coordinates.
		let ilm = theLatticeMapping.inverse	//	ilm = inverse lattice mapping
		var theMinCoords: SIMD3<Int> = .zero
		var theMaxCoords: SIMD3<Int> = .zero
		for c in 0...2 {	//	c says which coordinate: x, y or z
		
			//	Ignoring for a moment the translational parts,
			//	the inverse lattice mapping's columns give the dual basis
			//	for the original lattice mapping's rows.
			let theDualVector = SIMD3<Double>(ilm[0][c], ilm[1][c], ilm[2][c])
			
			//	If we rescale theDualVector to have length tilingRadiusÅ,
			//	it will give us the point on the tiling sphere
			//	whose c-coordinate back in the cubic space is maximal.
			let theDualVectorLength = length(theDualVector)
			precondition(
				theDualVectorLength > 0.0001,
				"Internal error: theDualVectorLength is unreasonably close to zero")
			let theMaxPhysicalPoint = (tilingRadiusÅ / theDualVectorLength) * theDualVector
			let theMinPhysicalPoint = -theMaxPhysicalPoint
			
			//	Find the preimages of the max and min physical points.
			let theMaxLatticePoint = ilm	//	right-to-left matrix action
								   * SIMD4<Double>(theMaxPhysicalPoint, 1.0)
			let theMinLatticePoint = ilm	//	right-to-left matrix action
								   * SIMD4<Double>(theMinPhysicalPoint, 1.0)

			theMaxCoords[c] = Int(floor(theMaxLatticePoint[c]))
			theMinCoords[c] = Int( ceil(theMinLatticePoint[c]))
		}
		
		//	If the tiling sphere's preimage falls entire between
		//	two adjecent integer coordinate plane in cubic space,
		//	then the tiling sphere will contain no visible instances
		//	of the given entity. Because Swift for-loops don't allow
		//	an empty range of values, we must explicitly test whether
		//	theMaxCoords[c] < theMinCoords[c] for any value of c.
		for c in 0...2 {
			if theMaxCoords[c] < theMinCoords[c] {
				//	Return an empty collection, which RealityKit
				//	treats as if it contained an identify transformation.
				return []
			}
		}

		var theInstances: [MeshResource.Instance] = []
		for i in theMinCoords[0] ... theMaxCoords[0] {
			for j in theMinCoords[1] ... theMaxCoords[1] {
				for k in theMinCoords[2] ... theMaxCoords[2] {
				
					let theCubicPoint = SIMD4<Double>(Double(i), Double(j), Double(k), 1.0)
					let thePhysicalPoint4 = theLatticeMapping * theCubicPoint
												//	right-to-left matrix action
					let thePhysicalPoint3 = SIMD3<Double>(
											thePhysicalPoint4.x,
											thePhysicalPoint4.y,
											thePhysicalPoint4.z)
					let theRadius = length(thePhysicalPoint3)
					
					if theRadius <= tilingRadiusÅ {
					
						//	Convert dimensions from ångströms in the crystal
						//	to meters in the user's physical space.
						let metersPerÅngström = 1e-10
						let theTranslationInMeters = scaleFactor
												   * metersPerÅngström
												   * thePhysicalPoint3
						
						//	Lift the whole crystal to center it near the user's eyes.
						//	Women's eye heights typically range from 144 to 165 cm
						//	above the floor, while men's range from 156 to 177 cm.
						//	So let's go with 160 cm, which falls in the overlap.
						let theEyeHeight = 1.6	//	meters
						let theOffset = SIMD3<Double>(0.0, theEyeHeight, 0.0)
						
						let theAdjustedTranslation
							= SIMD4<Double>(theTranslationInMeters + theOffset, 1.0)

						let theTranform = simd_float4x4(
							SIMD4<Float>(1.0, 0.0, 0.0, 0.0),
							SIMD4<Float>(0.0, 1.0, 0.0, 0.0),
							SIMD4<Float>(0.0, 0.0, 1.0, 0.0),
							SIMD4<Float>(theAdjustedTranslation) )
						
						let theBrightnessLevel: Int
						switch tilingRadiusÅ - theRadius {	//	how far from the boundary?
							case 0.0 ..< 4.0:	theBrightnessLevel = 3	//	25% brightness
							case 4.0 ..< 8.0:	theBrightnessLevel = 2	//	50% brightness
							case 8.0 ..< 12.0:	theBrightnessLevel = 1	//	75% brightness
							default:			theBrightnessLevel = 0	//	full brightness
						}
						let theInstance = MeshResource.Instance(
											id: "instance" + "-\(theInstances.count)",
											model: "model \(theBrightnessLevel)",
											at: theTranform)
						
						theInstances.append(theInstance)
					}
				}
			}
		}
		
		return theInstances
	}

	func makeMaterials(
		linearDisplayP3Color: SIMD3<Double>,
		roughness: Double,
		metallic: Double
	) -> [PhysicallyBasedMaterial] {

		//	Faux fog use a sequence of four dimming levels.
		let theDimmingLevels = [0,1,2,3]
		let theBrightness = [1.0, 0.75, 0.25, 0.0625]
		let theShininessFactor = [1.0, 0.875, 0.75, 0.0]

		let theMaterials = theDimmingLevels.map() { i in

			let theDimmedColor = theBrightness[i] * linearDisplayP3Color
			let theGammaEncodedColor = SIMD3<Double>(
				gammaEncode(theDimmedColor[0]),
				gammaEncode(theDimmedColor[1]),
				gammaEncode(theDimmedColor[2]))
			let theUIColor = UIColor(
								displayP3Red: theGammaEncodedColor[0],
								green: theGammaEncodedColor[1],
								blue: theGammaEncodedColor[2],
								alpha: 1.0)
			
			let s = theShininessFactor[i]
			let theRoughness = Float(1.0 - s * (1.0 - roughness)) // roughess = 1.0 - shininess
			let theMetallic  = Float(s * metallic)

			var theMaterial = PhysicallyBasedMaterial()
			theMaterial.baseColor = .init(tint: theUIColor)
			theMaterial.roughness = .init(floatLiteral: theRoughness)
			theMaterial.metallic  = .init(floatLiteral: theMetallic)
			if i == 3 {
				theMaterial.sheen = .init(tint: .black)
			}
			
			return theMaterial
		}
		
		return theMaterials
	}
	

// MARK: -
// MARK: Tap gesture

#warning("Remove this doubleTapGesture in a future version, when user will know to press the Digital Crown to dismiss a crystal.")
	var doubleTapGesture: some Gesture {
		TapGesture(count: 2)
		.targetedToAnyEntity()
		.onEnded() { value in
			openWindow(id: "TEMPORARY alert", value: 0)
		}
	}
	

// MARK: -
// MARK: Magnify gesture

	@State var atomicRadiusFactorAtGestureStart: Double? = nil

	var atomMagnifyGesture: some Gesture {
	
		//	Note: In the simulator, MagnifyGesture() responds only
		//	to horizontal "magnify" motions, pulling your hands apart
		//	in any other direction gives surprising results. But on
		//	the actual device, MagnifyGesture() works correctly in any direction.
		
		MagnifyGesture()
		.targetedToAnyEntity()
		.onChanged() { value in

			//	When the MagnifyGesture begins,
			//	note the current value of the atomicRadiusFactor.
			if atomicRadiusFactorAtGestureStart == nil {
				atomicRadiusFactorAtGestureStart = atomicRadiusFactor
			}
			
			guard let theValueAtStart = atomicRadiusFactorAtGestureStart else { return }
			
			let theRawValue = theValueAtStart * value.magnification
			let theClampedValue = max(
									min(
										theRawValue,
										maxAtomicRadiusFactor),
									minAtomicRadiusFactor)
			
			atomicRadiusFactor = theClampedValue
		}
		.onEnded() { value in

			//	When the MagnifyGesture ends,
			//	clear the atomicRadiusFactor's starting value.
			atomicRadiusFactorAtGestureStart = nil
		}
	}
}
