//	GGVR_GPU_Headers.h
//
//	© 2024 by Jeff Weeks
//	See TermsOfUse.txt

#ifndef GGVR_GPU_Headers_h
#define GGVR_GPU_Headers_h

#include <simd/simd.h>

//	Define shared GPU functions right here in this .h file,
//	so the compiler can inline them into whatever app-specific
//	GPU function calls them. (In other words, make sure they
//	end up in the same "compilation unit".)
//
//		Note: The compiler would probably inline them
//		with or without the 'inline' keyword. We include it
//		not only to make our intentions clear, but also
//		to avoid a linker error in the unlikely event that
//		we ever include this .h file in than one .metal file
//		in the same project.
//

inline float4 azimuthalEquidistantMap(
	float4	positionInBridgeOfNoseSpace,
	uchar	geometryCode,	//	I couldn't figure out how to share function constants
	float	spaceRadius,	//		among different .metal files,
	float	spaceRadiusInv)	//		so I passed them as parameters instead.
{
	float	factor = 1.0;	//	overwritten below
	float4	positionInTangentSpace;

	//	Given that the positionInBridgeOfNoseSpace sits
	//	at some distance r from the player (as measured
	//	in the non-Euclidean space itself), in some direction v,
	//	we want to find a point in the Euclidean tangent space
	//	(based at the bridge of the player's nose) that
	//	sits in that same direction v at that same distance r,
	//	but in the Euclidean tangent space rather than
	//	in the non-Euclidean space itself.
	//
	//	Let's compute the positionInTangentSpace as explained and
	//	illustrated in §5 of
	//
	//		Jeff Weeks,
	//		Body coherence in curved-space virtual reality games,
	//		Computers & Graphics 97 (2021) 28–41.
	//
	//		  https://www.sciencedirect.com/science/article/pii/S0097849321000443
	//		or
	//		  https://arxiv.org/abs/2011.00510
	//
	
	if (geometryCode == 's')
	{
		float cosTheta = clamp(
							positionInBridgeOfNoseSpace.w * spaceRadiusInv,
							-1.0,
							+1.0);
		float theta = acos(cosTheta);
		float sinTheta = sin(theta);
		
		if (sinTheta > 0.000001)	//	typical case
		{
			factor = theta / sinTheta;
			positionInTangentSpace = float4(
					factor * positionInBridgeOfNoseSpace.x,
					factor * positionInBridgeOfNoseSpace.y,
					factor * positionInBridgeOfNoseSpace.z,
					spaceRadius);
		}
		else	//	sin(θ) ≤ 0.000001
		{
			if (cosTheta > 0.0)	//	near north pole, cos(θ) ≈ +1.0
			{
				//	Near the north pole, the angle θ is small and the difference
				//
				//		θ - sin(θ) = θ - (θ - θ³/6 + θ⁵/120 - …)
				//				   = θ³/6 - θ⁵/120 + …
				//				   ≈ θ³/6
				//
				//	is much much smaller.  So the "factor" computed above
				//	is extremely close to 1.0 and the positionInTangentSpace
				//	is extremely close to the positionInBridgeOfNoseSpace.
				//	Note that this conclusion remains valid even when
				//	the vertex sits exactly at the north pole,
				//	where θ = 0.0 and the factor = θ/sin(θ) = “0.0/0.0” = 1.0.
				
				positionInTangentSpace = float4(
						positionInBridgeOfNoseSpace.x,
						positionInBridgeOfNoseSpace.y,
						positionInBridgeOfNoseSpace.z,
						spaceRadius);	//	w coordinate almost same as before
			}
			else	//	near south pole, cos(θ) ≈ -1.0
			{
				//	At the south pole the situation is a bit trickier:
				//	we know that the south pole itself should map to a point
				//	exactly (π * spaceRadius) meters from the observer,
				//	but its direction is completely undefined.
				//	Nevertheless, we have to send the south pole somewhere,
				//	so let's put it directly in front of the player at
				//
				//			(0, 0, π * spaceRadius, spaceRadius)
				//
				positionInTangentSpace = float4(
						0.0,
						0.0,
						M_PI_F * spaceRadius,
						spaceRadius);
			}
		}
	}

	if (geometryCode == 'e')
	{
		positionInTangentSpace = positionInBridgeOfNoseSpace;
	}

	if (geometryCode == 'h')
	{
		float coshTheta = max(
							positionInBridgeOfNoseSpace.w * spaceRadiusInv,
							+1.0);
		float theta = acosh(coshTheta);
		float sinhTheta = sinh(theta);
		factor = (sinhTheta > 0.000001 ?
					theta / sinhTheta :	//	typical case
					1.0);				//	near north pole
		positionInTangentSpace = float4(
				factor * positionInBridgeOfNoseSpace.x,
				factor * positionInBridgeOfNoseSpace.y,
				factor * positionInBridgeOfNoseSpace.z,
				spaceRadius);
	}

	return positionInTangentSpace;
}

inline float4 compressSouthernHemisphere(	//	returns positionInSafeBridgeOfNoseSpace
	float4	positionInBridgeOfNoseSpace,
	float	spaceRadius)
{
	float3	xyz;
	float	w,
			lengthXYZ,
			newLengthXYZ;
	float4	positionInSafeBridgeOfNoseSpace;

	const float antipodeEpsilon = 0.001;	//	1 mm
	const float minAcceptableW = -(spaceRadius - antipodeEpsilon);
	const float outerHorizonWInRadians = 0.001;
	const float outerHorizonW = outerHorizonWInRadians * spaceRadius;	//	in meters
	const float innerHorizonWInRadians = 0.002;
	const float innerHorizonW = innerHorizonWInRadians * spaceRadius;	//	in meters
	const float wCompressionFactor = (innerHorizonW -  outerHorizonW )
								   / (innerHorizonW - (-spaceRadius));

	xyz = positionInBridgeOfNoseSpace.xyz;
	w   = positionInBridgeOfNoseSpace.w;
	
	//	Rendering the antipodal point (0, 0, 0, -gSpaceRadius) is tricky,
	//	because it's equally far away in all directions. So before doing
	//	anything else, let's test whether positionInBridgeOfNoseSpace
	//	is at or near the antipodal point and, if so, nudge it a tiny bit
	//	so it sits straight in front of us.  For full robustness, let's
	//	make sure xyz and w both fall within the acceptable range.
	
	lengthXYZ = length(xyz);
	if (lengthXYZ < antipodeEpsilon)
	{
		xyz = float3(0.0, 0.0, -antipodeEpsilon);
		lengthXYZ = antipodeEpsilon;
	}
	
	if (w < minAcceptableW)
	{
		w = minAcceptableW;
	}

	//	Project the whole southern hemisphere (except the south pole,
	//	which the preceding code tweaks to a safe location)
	//	into a spherical shell just north of the equatorial 2-sphere.
	if (w < innerHorizonW)
	{
		//	Map w values in the range -gSpaceRadius < w < +innerHorizonW
		//	linearly onto the range +outerHorizonW < w < +innerHorizonW.
		w = innerHorizonW - (innerHorizonW - w) * wCompressionFactor;
		
		//	Adjust the length of the xyz part accordingly.
		newLengthXYZ = sqrt( spaceRadius*spaceRadius - w*w );
		xyz *= newLengthXYZ / lengthXYZ;
	}
	
	positionInSafeBridgeOfNoseSpace = float4(xyz, w);
		
	return positionInSafeBridgeOfNoseSpace;
}

#endif /* GGVR_GPU_Headers_h */
